import contextlib
from datetime import datetime, timedelta
import uuid
from typing import Dict, Union

from fastapi.encoders import jsonable_encoder
from httpx import AsyncClient
from hypothesis import given, strategies as st, settings
import pytest
import sqlalchemy as sa

from c3loc.api2.main import app
from c3loc.api2.schemas.tags import TagPatch, IBeacon, SecureSmartRelay
from c3loc.api2.schemas.sensors import SensorType
from c3loc.config import CONFIG
from c3loc.db_next import history, sensors

from .test_api2_groups import temporary_group
from .test_api2_zones import temporary_zone

@pytest.fixture
def delete_tag():
    async def _del(tag_id: int) -> None:
        async with AsyncClient(app=app, base_url='http://test') as client:
            await client.delete(f'/tags/{tag_id}')
    return _del


@pytest.fixture
def new_sr_tag():
    async def _add() -> int:
        async with AsyncClient(app=app, base_url='http://test') as client:
            response = await client.post('/tags', json={
                'mac': '00:00:00:00:00:00',
                'type': 'SmartRelay',
                'name': 'Test Tag'})
        assert response.status_code == 201
        tag_id = int(response.headers['Location'].split('/')[-1])
        return tag_id
    return _add


@contextlib.asynccontextmanager
async def temporary_sr_tag():
    print("Creating SR")
    async with AsyncClient(app=app, base_url='http://test') as client:
        response = await client.post('/tags', json={
            'mac': '00:00:00:00:00:00',
            'type': 'SmartRelay',
            'name': 'Test Tag'})
    assert response.status_code == 201
    tag_id = int(response.headers['Location'].split('/')[-1])
    try:
        yield tag_id
    finally:
        print("Destroying SR")
        async with AsyncClient(app=app, base_url='http://test') as client:
            await client.delete(f'/tags/{tag_id}')


@pytest.fixture
async def mock_history(engine):
    async with engine.connect() as conn:
        async with temporary_zone() as zone_id:
            async with temporary_sr_tag() as tag_id:
                insert = history.insert().values({
                    'tag_id': tag_id,
                    'zone_id': zone_id,
                    'start_ts': datetime.utcnow(),
                    'distance': 0})
                r = await conn.execute(insert)
                hist_id = r.inserted_primary_key[0]
                await conn.commit()
                yield tag_id
                await conn.execute(
                    history.delete().where(history.c.id == hist_id))
                await conn.commit()


@pytest.fixture
async def mock_contact(engine):
    async with engine.connect() as conn:
        async with temporary_sr_tag() as main_tag:
            async with temporary_zone() as zone_id:
                insert = history.insert().values({
                    'tag_id': main_tag,
                    'zone_id': zone_id,
                    'start_ts': datetime.utcnow(),
                    'distance': 0,
                    'zone_name': 'Farts'})
                await conn.execute(insert)
                async with temporary_sr_tag() as other_tag:
                    insert = history.insert().values({
                        'tag_id': other_tag,
                        'zone_id': zone_id,
                        'start_ts': datetime.utcnow(),
                        'distance': 0,
                        'zone_name': "Farts"})
                    await conn.execute(insert)
                    await conn.commit()
                    yield main_tag, other_tag
                    await conn.execute(history.delete())
                    await conn.commit()


@pytest.fixture
async def mock_no_contact(engine):
    async with engine.connect() as conn:
        async with temporary_sr_tag() as main_tag:
            async with temporary_zone() as zone_id:
                insert = history.insert().values({
                    'tag_id': main_tag,
                    'zone_id': zone_id,
                    'start_ts': datetime.utcnow()-timedelta(
                        minutes=10),
                    'end_ts': datetime.utcnow()-timedelta(
                        minutes=5),
                    'distance': 0,
                    'zone_name': 'Farts'})
                await conn.execute(insert)
                async with temporary_sr_tag() as other_tag:
                    insert = history.insert().values({
                        'tag_id': other_tag,
                        'zone_id': zone_id,
                        'start_ts': datetime.utcnow(),
                        'distance': 0,
                        'zone_name': "Farts"})
                    await conn.execute(insert)
                    await conn.commit()
                    yield main_tag, other_tag
                    await conn.execute(history.delete())
                    await conn.commit()


@pytest.fixture
async def mock_sensor(engine):
    async with engine.connect() as conn:
        async with temporary_sr_tag() as tag_id:
            insert = sensors.insert().values({
                'tag_id': tag_id,
                'type': SensorType.SensorTempData,
                'value': b'\x00'})
            r = await conn.execute(insert)
            sensor_id = r.inserted_primary_key[0]
            await conn.commit()
            yield tag_id
            await conn.execute(
                sensors.delete().where(sensors.c.id == sensor_id))
            await conn.commit()


@pytest.mark.asyncio
async def test_get_contacts(mock_contact, client):
    main_tag, other_tag = mock_contact
    response = await client.get(f'/tags/{main_tag}/contacts')
    assert response.status_code == 200
    body = response.json()
    assert len(body) == 1
    assert body[0]['tag_id'] == other_tag

    response = await client.get(f'/tags/{main_tag}/contacts?offset=1')
    assert response.status_code == 200
    assert len(response.json()) == 0

    response = await client.get(f'/tags/{main_tag}/contacts?limit=0')
    assert response.status_code == 200
    assert len(response.json()) == 0


@pytest.mark.asyncio
async def test_get_contacts_exclude_non_contact(mock_no_contact, client):
    main_tag, other_tag = mock_no_contact
    response = await client.get(f'/tags/{main_tag}/contacts')
    assert response.status_code == 200
    body = response.json()
    assert len(body) == 0


@pytest.mark.asyncio
async def test_get_contacts_404(client):
    response = await client.get(f'/tags/1000/contacts')
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_get_contacts_time_filter(client, mock_contact):
    main_tag, _ = mock_contact
    response = await client.get(
            f'/tags/{main_tag}/contacts?start={datetime.utcnow().isoformat()}'
            f'&end={datetime.utcnow().isoformat()}')
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_get_history(mock_history, client):
    new_id = mock_history
    response = await client.get(f'/tags/{new_id}/history')
    assert response.status_code == 200
    assert len(response.json()) == 1

    response = await client.get(f'/tags/{new_id}/history?offset=1')
    assert response.status_code == 200
    assert len(response.json()) == 0

    response = await client.get(f'/tags/{new_id}/history?limit=0')
    assert response.status_code == 200
    assert len(response.json()) == 0


@pytest.mark.asyncio
async def test_get_history_404(client):
    response = await client.get(f'/tags/1000/history')
    assert response.status_code == 404


@pytest.mark.asyncio
async def test_get_history_time_filter(client):
    async with temporary_sr_tag() as tag_id:
        response = await client.get(
            f'/tags/{tag_id}/history?start={datetime.utcnow().isoformat()}'
            f'&end={datetime.utcnow().isoformat()}')
        assert response.status_code == 200


@pytest.mark.asyncio
async def test_get_history_time_filter(client):
    async with temporary_sr_tag() as tag_id:
        response = await client.get(
            f'/tags/{tag_id}/history?start={datetime.utcnow().isoformat()}'
            f'&end={datetime.utcnow().isoformat()}')
        assert response.status_code == 200


@pytest.mark.asyncio
async def test_get_history_time_filter_bad_iso(client):
    async with temporary_sr_tag() as tag_id:
        response = await client.get(f'/tags/{tag_id}/history?start=11')
        assert response.status_code == 422


@pytest.mark.asyncio
async def test_get_tags_null(client):
    response = await client.get("/tags")
    assert response.status_code == 200
    assert response.json() == []


@pytest.mark.asyncio
async def test_get_tags(client):
    async with temporary_sr_tag() as tag_id:
        response = await client.get(f'/tags')
        assert response.status_code == 200
        ts = response.json()
        assert len(ts) == 1
        assert ts[0]['id'] == tag_id

        response = await client.get('/tags?offset=1')
        assert response.status_code == 200
        ts = response.json()
        assert len(ts) == 0

        response = await client.get('/tags?limit=0')
        assert response.status_code == 200
        ts = response.json()
        assert len(ts) == 0


@pytest.mark.asyncio
async def test_tags_filter_type():
    async with temporary_sr_tag() as new_id:
        async with AsyncClient(app=app, base_url='http://test') as client:
            response = await client.get('/tags?type=SmartRelay')
            assert response.status_code == 200
            assert response.json()[0]['id'] == new_id

            response = await client.get('/tags?type=iBeacon')
            print('Response', response.json())
            assert response.status_code == 200
            assert len(response.json()) == 0


@pytest.mark.asyncio
async def test_tags_filter_group():
    async with temporary_group() as group_id:
        async with temporary_sr_tag() as new_id:
            async with AsyncClient(app=app, base_url='http://test') as client:
                response = await client.patch(f'/tags/{new_id}', json={
                    'group_id': group_id
                })
                assert response.status_code == 204

                response = await client.get(f'/tags?group_id={group_id}')
                assert response.status_code == 200
                assert response.json()[0]['id'] == new_id

                response = await client.get(f'/tags?group_id={group_id+1}')
                assert response.status_code == 200
                assert len(response.json()) == 0


@pytest.mark.asyncio
async def test_delete(new_sr_tag, client):
    new_id = await new_sr_tag()
    response = await client.delete(f'/tags/{new_id}')
    assert response.status_code == 204


@pytest.mark.asyncio
async def test_tags_post_bad_no_type(client):
    tag = {'name': 'Test', 'mac': '00:00:00:00:00:00'}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_sr_bad_id_null(client):
    tag = {'name': 'Test', 'type': 'SmartRelay'}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_sr_bad_id_bid(client):
    tag = {'name': 'Test', 'type': 'SmartRelay', 'bid': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_sr_bad_id_uuid(client):
    tag = {'name': 'Test', 'type': 'SmartRelay',
           'uuid': 'f03b1170-9f91-4e10-b640-136d317d3228',
           'major': 0, 'minor': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_sr_bad_id_bid_out_of_range(client):
    tag = {'name': 'Test', 'type': 'SmartRelay',
           'bid': 2**32}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422

    tag = {'name': 'Test', 'type': 'SmartRelay',
           'bid': -1}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_sr_bad_mac(client):
    tag = {'name': 'Test', 'type': 'SmartRelay', 'mac': '00:00'}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_sr_default_name(client):
    tag = {'type': 'SmartRelay', 'mac': '00:00:00:00:00:00'}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 201
    new_id = int(response.headers['Location'].split('/')[-1])
    response = await client.get(f'/tags/{new_id}')
    assert response.status_code == 200
    name = response.json()['name']
    assert name


def format_mac(mac_bytes: bytes) -> str:
    return ':'.join([f'{x:02x}' for x in range(len(mac_bytes))])


@given(name=st.text(alphabet=st.characters(blacklist_categories=('Cc', 'Cs'))),
       mac_addr=st.builds(format_mac, st.binary(min_size=6, max_size=6)),
       attribs=st.one_of(
           st.none(),
           st.from_type(Dict[str, Union[str, int]])))
@settings(deadline=None)
@pytest.mark.asyncio
async def test_post_sr(name, mac_addr, attribs, engine):
    async with temporary_group() as group_id:
        async with temporary_zone() as zone_id:
            async with AsyncClient(app=app, base_url='http://test') as client:
                response = await client.post("/tags", json={
                    'name': name,
                    'mac': mac_addr,
                    'type': 'SmartRelay',
                    'attributes': attribs,
                    'group_id': group_id,
                    'zone_id': zone_id
                })
            assert response.status_code == 201
            async with engine.connect() as conn:
                await conn.execute(sa.text('delete from tags'))
                await conn.commit()


@pytest.mark.asyncio
async def test_tags_post_ssr_bad_id_null(client):
    tag = {'name': 'Test', 'type': 'SecureSmartRelay'}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ssr_bad_id_mac(client):
    tag = {'name': 'Test', 'type': 'SecureSmartRelay',
           'mac': '00:00:00:00:00:00'}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ssr_bad_id_uuid(client):
    tag = {'name': 'Test', 'type': 'SecureSmartRelay',
           'uuid': 'f03b1170-9f91-4e10-b640-136d317d3228',
           'major': 0, 'minor': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ssr_bad_bid(client):
    tag = {'name': 'Test', 'type': 'SecureSmartRelay', 'bid': 2**32}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422

    tag = {'name': 'Test', 'type': 'SecureSmartRelay', 'bid': -1}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_ssr_default_name(client):
    tag = {'type': 'SecureSmartRelay', 'bid': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 201
    new_id = int(response.headers['Location'].split('/')[-1])
    response = await client.get(f'/tags/{new_id}')
    assert response.status_code == 200
    name = response.json()['name']
    assert name


@given(ssr=st.builds(SecureSmartRelay))
@settings(deadline=None)
@pytest.mark.asyncio
async def test_post_ssr(ssr, engine):
    async with AsyncClient(app=app, base_url='http://test') as client:
        response = await client.post("/tags", json=ssr.dict())
    assert response.status_code == 201
    async with engine.connect() as conn:
        await conn.execute(sa.text('delete from tags'))


@pytest.mark.asyncio
async def test_tags_post_ib_bad_id_null(client):
    tag = {'name': 'Test', 'type': 'iBeacon'}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ib_bad_id_mac(client):
    tag = {'name': 'Test', 'type': 'iBeacon',
           'mac': '00:00:00:00:00:00'}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ib_bad_id_bid(client):
    tag = {'name': 'Test', 'type': 'iBeacon', 'bid': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ib_bad_id_bad_major(client):
    tag = {'name': 'Test', 'type': 'iBeacon',
           'uuid': str(uuid.uuid4()), 'major': 2**16, 'minor': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422

    tag = {'name': 'Test', 'type': 'iBeacon',
           'uuid': str(uuid.uuid4()), 'major': -1, 'minor': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ib_bad_id_bad_minor(client):
    tag = {'name': 'Test', 'type': 'iBeacon',
           'uuid': str(uuid.uuid4()), 'major': 0, 'minor': 2**16}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422

    tag = {'name': 'Test', 'type': 'iBeacon',
           'uuid': str(uuid.uuid4()), 'major': 0, 'minor': -1}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ib_bad_uuid(client):
    tag = {'name': 'Test', 'type': 'iBeacon',
     'uuid': 'f03b1170-9f91', 'major': 0, 'minor': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ib_missing_major(client):
    tag = {'name': 'Test', 'type': 'iBeacon',
           'uuid': 'f03b1170-9f91-4e10-b640-136d317d3228',
           'minor': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_post_ib_missing_minor(client):
    tag = {'name': 'Test', 'type': 'iBeacon',
           'uuid': 'f03b1170-9f91-4e10-b640-136d317d3228',
           'major': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_tags_ib_default_name(client):
    tag = {'type': 'iBeacon',
           'uuid': 'f03b1170-9f91-4e10-b640-136d317d3228',
           'major': 0, 'minor': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 201
    new_id = int(response.headers['Location'].split('/')[-1])
    response = await client.get(f'/tags/{new_id}')
    assert response.status_code == 200
    name = response.json()['name']
    assert name


@pytest.mark.asyncio
async def test_tags_la_default_name(client):
    tag = {'type': 'iBeacon',
           'uuid': str(CONFIG['LA_UUID']),
           'major': 0, 'minor': 0}
    response = await client.post("/tags", json=tag)
    assert response.status_code == 201
    new_id = int(response.headers['Location'].split('/')[-1])
    response = await client.get(f'/tags/{new_id}')
    assert response.status_code == 200
    name = response.json()['name']
    assert name


@given(ib=st.builds(IBeacon))
@settings(report_multiple_bugs=False, deadline=None)
@pytest.mark.asyncio
async def test_post_ib(ib, engine):
    async with AsyncClient(app=app, base_url='http://test') as client:
        response = await client.post("/tags", json=jsonable_encoder(ib))
        assert response.status_code == 201
    async with engine.connect() as conn:
        await conn.execute(sa.text('delete from tags'))


@given(patch=st.builds(TagPatch))
@pytest.mark.asyncio
async def test_patch_tag(patch):
    async with temporary_sr_tag() as new_id:
        async with AsyncClient(app=app, base_url='http://test') as client:
            response = await client.patch(f'/tags/{new_id}', json=patch.dict())
            assert response.status_code == 204
            response = await client.get(f'/tags/{new_id}')
            assert response.status_code == 200
            body = response.json()
            for k, v in patch.dict().items():
                assert body[k] == patch.dict()[k]


@pytest.mark.asyncio
async def test_tags_404(client):
    response = await client.get('/tags/1000')
    assert response.status_code == 404
    response = await client.patch('/tags/1000', json={'name': 'farts'})
    assert response.status_code == 404
    response = await client.delete('/tags/1000')
    assert response.status_code == 404



from datetime import datetime
from typing import Optional

from pydantic import BaseModel  # type: ignore

from .types import DbInteger
from .tags import TagType


class Proximity(BaseModel):
    tag_id: DbInteger
    tag_name: str
    zone_name: Optional[str]
    last_seen: datetime
    tag_type: TagType
    distance: Optional[float]
    alarm_active: Optional[bool]



"""Adds alarm_active field to tags

Revision ID: 720d3e9a9af0
Revises: fa1582c9a701
Create Date: 2021-02-10 16:08:32.143061

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '720d3e9a9af0'
down_revision = 'fa1582c9a701'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("ALTER TABLE tags ADD COLUMN alarm_active bool;")


def downgrade():
    op.execute("ALTER TABLE tags DROP COLUMN alarm_active;")



from aiohttp import web

from .schemas import alarms_schema, alarm_schema, alarm_patch_schema
from .views import ValidatingView, paginate_query


def process_alarm_links(result):
    zone_id = result['zone_id']
    tag_id = result['tag_id']
    d_r = dict(result)
    d_r['links'] = {
        "tag": f"/api/tags/{tag_id}",
        "zone": f"/api/zones/{zone_id}" if zone_id else None
    }
    return d_r


class AlarmsView(ValidatingView):
    async def get(self):
        async with self.request.app['db_pool'].acquire() as conn:
            query = None
            if 'acknowledged' in self.request.query:
                try:
                    ack_filter = int(self.request.query['acknowledged'])
                except ValueError:
                    raise web.HTTPBadRequest(
                        reason="Invalid value for query param 'acknowledged' "
                               "(should be 0 or 1)")
                query = (
                    'SELECT alarms.id as id, tag_id, tags.name as tag_name, '
                    'zones.name as zone_name, start_ts, last_ts, '
                    'acknowledged, zone_id, priority '
                    'FROM alarms JOIN tags ON tag_id = tags.id '
                    'LEFT JOIN zones ON tags.zone_id = zones.id '
                    'WHERE acknowledged = $1', bool(ack_filter))
            else:
                query = (
                    'SELECT alarms.id as id, tag_id, tags.name as tag_name, '
                    'zones.name as zone_name, start_ts, last_ts, '
                    'acknowledged, zone_id, priority '
                    'FROM alarms JOIN tags ON tag_id = tags.id '
                    'LEFT JOIN zones ON tags.zone_id = zones.id', )
            alarms = await conn.fetch(*paginate_query(self.request, query))
            return web.json_response(alarms_schema.dump(
                [process_alarm_links(r) for r in alarms]
            ))


class AlarmView(ValidatingView):
    async def get(self):
        a_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            alarm = await conn.fetchrow(
                'SELECT alarms.id as id, tag_id, tags.name, '
                'zones.name as zone_name, start_ts, last_ts, '
                'acknowledged as tag_name, zone_id, priority '
                'FROM alarms JOIN tags ON tag_id = tags.id '
                'LEFT JOIN zones ON tags.zone_id = zones.id '
                'WHERE alarms.id = $1', a_id)
            if not alarm:
                raise web.HTTPNotFound
            return web.json_response(
                alarm_schema.dump(process_alarm_links(alarm)))

    async def patch(self):
        a_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            l_row = await conn.fetchrow(
                'SELECT acknowledged FROM alarms WHERE id = $1', a_id)
            if not l_row:
                raise web.HTTPNotFound
            body = await self._valid_json(self.request)
            alarm = alarm_patch_schema.load(body)
            if alarm['acknowledged']:
                await conn.execute(
                    'UPDATE alarms SET ack_ts = now() WHERE id = $1', a_id)
            else:
                await conn.execute(
                    'UPDATE alarms SET ack_ts = NULL WHERE id = $1', a_id)
        raise web.HTTPNoContent

    async def delete(self):
        a_id = int(self.request.match_info['id'])
        async with self.request.app['db_pool'].acquire() as conn:
            a_row = await conn.fetchrow(
                'SELECT id FROM alarms WHERE id = $1', a_id)
            if not a_row:
                raise web.HTTPNotFound
            await conn.execute('DELETE from alarms where id = $1', a_id)
        raise web.HTTPNoContent



import asyncpg  # type: ignore

from .config import CONFIG, DB_RELEVANT


async def init_db_pool(app_name):
    db_url = f"postgres://{CONFIG['DB_USER']}:{CONFIG['DB_PASSWORD']}@{CONFIG['DB_HOST']}/{CONFIG['DB_NAME']}"   # noqa: E501
    db_url_redacted = f"postgres://{CONFIG['DB_USER']}:***@{CONFIG['DB_HOST']}"
    print(f"Connecting to {db_url_redacted}")
    pool = await asyncpg.create_pool(
        dsn=db_url, min_size=CONFIG['MAX_DB_CONNECTIONS'],
        max_size=CONFIG['MAX_DB_CONNECTIONS'])
    async with pool.acquire() as conn:
        for k, v in DB_RELEVANT[app_name].items():
            await conn.execute(
                'INSERT INTO settings (name, value) VALUES ($1,$2) '
                'ON CONFLICT ON CONSTRAINT settings_pkey DO '
                'UPDATE SET value = $2 WHERE settings.name = $1;', k, str(v))
    return pool



import json
import string

from marshmallow import Schema, fields


class JsonField(fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        if value:
            return json.loads(value)
        return {}

    def _deserialize(self, value, attr, data, **kwargs):
        return value


class AttrsMixin:
    attrs = JsonField(allow_none=True, default={})


class ListenerSchema(Schema, AttrsMixin):
    id = fields.Str(dump_only=True)
    name = fields.Str(allow_none=True)
    zone_id = fields.Int(allow_none=True)
    last_seen = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', dump_only=True)


class ZoneSchema(Schema, AttrsMixin):
    id = fields.Int(dump_only=True)
    name = fields.Str()


class GroupSchema(Schema, AttrsMixin):
    id = fields.Int(dump_only=True)
    name = fields.Str()


def validate_mac(value):
    valid_char = set(string.hexdigits)
    valid_char.add(':')
    return all(c in set(valid_char) for c in value)


class MacMixin:
    mac = fields.Str(validate=validate_mac)


class RequiredMacMixin:
    mac = fields.Str(required=True, validate=validate_mac)


class TagPatchSchema(Schema, AttrsMixin):
    name = fields.Str(allow_none=True)
    zone_id = fields.Int(allow_none=True, validate=lambda x: x > 0)
    group_id = fields.Int(allow_none=True, validate=lambda x: x > 0)


class TagCreateBase(Schema, AttrsMixin):
    id = fields.Int(dump_only=True)
    name = fields.Str(allow_none=True)
    zone_id = fields.Int(allow_none=True)
    group_id = fields.Int(allow_none=True)
    last_seen = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', dump_only=True)
    type = fields.Str(required=True, validate=lambda x: x in {
        'iBeacon', 'SmartRelay', 'LocationAnchor', 'SecureLocationAnchor',
        'SecureSmartRelay', 'MacBeacon'})
    battery_pct = fields.Int(allow_none=True)
    alarm_active = fields.Boolean(allow_none=True)
    zone_start = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', dump_only=True)


class iBeaconTagMixin:
    uuid = fields.UUID(required=True)
    major = fields.Int(required=True, validate=lambda x: x >= 0 < (2 ** 16))
    minor = fields.Int(required=True, validate=lambda x: x >= 0 < (2 ** 16))


class SecureBeaconTagMixin:
    bid = fields.Int(required=True)


class MacBeaconTagMixin(RequiredMacMixin):
    pass


class SecureBeaconCreateSchema(TagCreateBase, SecureBeaconTagMixin):
    pass


class iBeaconCreateSchema(TagCreateBase, iBeaconTagMixin):
    pass


class MacBeaconCreateSchema(TagCreateBase, MacBeaconTagMixin):
    pass


class ProxLinks(Schema):
    tag = fields.URL(required=True)
    zone = fields.URL(required=True)


class ProxSchema(Schema):
    tag_id = fields.Int(required=True)
    tag_name = fields.Str(required=True)
    distance = fields.Float(required=True)
    zone_name = fields.Str(required=True)
    last_seen = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', required=True)
    tag_type = fields.Str(required=True)
    alarm_active = fields.Str(required=True)
    links = fields.Nested(ProxLinks)


class AlarmLinks(Schema):
    tag = fields.URL(required=True)
    zone = fields.URL(required=True)


class AlarmSchema(Schema):
    id = fields.Int(required=True)
    tag_name = fields.Str(required=True)
    zone_name = fields.Str(required=True)
    start_ts = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', required=True)
    last_ts = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', required=True)
    acknowledged = fields.Boolean(required=True)
    priority = fields.Int(required=True)
    links = fields.Nested(AlarmLinks)


class AlarmPatch(Schema):
    acknowledged = fields.Boolean(required=True)


class HistorySchema(Schema):
    start_ts = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', required=True)
    end_ts = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', required=True)
    zone_id = fields.Int(required=True)
    zone_name = fields.Str(required=True)
    distance = fields.Float(required=True)
    duration = fields.Str(required=True)


class ContactsSchema(Schema):
    tag_id = fields.Int(required=True)
    tag_name = fields.Str(required=True)
    start_ts = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', required=True)
    end_ts = fields.DateTime('%Y-%m-%dT%H:%M:%S+00:00', required=True)
    duration = fields.TimeDelta(required=True)
    zone_id = fields.Int(required=True)
    zone_name = fields.Str(required=True)
    hash = fields.Str(required=True)


listeners_schema = ListenerSchema(many=True)
listener_schema = ListenerSchema()

zones_schema = ZoneSchema(many=True)
zone_schema = ZoneSchema()
groups_schema = GroupSchema(many=True)
group_schema = GroupSchema()

ibeacon_post_schema = iBeaconCreateSchema()
macbeacon_post_schema = MacBeaconCreateSchema()
secure_beacon_post_schema = SecureBeaconCreateSchema()
tag_patch_schema = TagPatchSchema()

prox_schema = ProxSchema(many=True)

alarm_schema = AlarmSchema()
alarms_schema = AlarmSchema(many=True)
alarm_patch_schema = AlarmPatch()

history_schema = HistorySchema(many=True)
contacts_schema = ContactsSchema(many=True)
